import glob
import os
import re

BLANK_RE_MATCH = r"^\s*$"
COMMENT_RE_MATCH = r"^\s*#(.*)$"
INCLUDE_RE_MATCH = r"^\s*((Include)|(IncludeOptional))\s+(.+)$"
DEFINE_RE_MATCH = r"Define\s+(\w+)\s+(.+)"
SERVER_ROOT_RE_MATCH = r"ServerRoot\s+(.+)"
SIMPLE_NODE_RE_MATCH = r"^\s*[a-zA-Z]\w*(\s+.*)*\s*[\\]?$"
COMPLEX_NODE_RE_MATCH = r"^\s*<\s*[a-zA-Z]\w*(\s+[^>]*)*\s*(>\s*|[\\])$"
COMPLEX_NODE_TAIL_RE_MATCH = r"\s*</%s>\s*$"
LEFT_TIP_SCRATCH = "<"
RIGHT_TIP_SCRATCH = ">"


class BaseNode:
    def __init__(self):
        self._lines = []
        self._finish_status = False
        self._fix_status = False
        self._name = ""

    def get_finish_status(self):
        return self._finish_status

    def get_fix_status(self):
        return self._fix_status

    def get_name(self):
        return self._name

    def get_lines(self):
        return self._lines

    def set_finish_status(self, val):
        self._finish_status = val

    def set_fix_status(self, val):
        self._fix_status = val

    def set_name(self, val):
        self._name = val

    def append_lines(self, line_content):
        self._fix_status = False
        new_line_content = line_content.lstrip("\\")
        if line_content == new_line_content:
            self._fix_status = True

        new_line_content = new_line_content.strip()
        new_content_list = new_line_content.split()
        self._name = new_content_list[0]
        value_contents = new_content_list[1:]
        for value_content in value_contents:
            self._lines.append(value_content)


class SimpleNode(BaseNode):
    def __init__(self):
        BaseNode.__init__(self)

    @property
    def arguments(self):
        return self._lines

    def append_lines(self, line_content):
        BaseNode.append_lines(self, line_content)


class HeaderNode(BaseNode):
    def __init__(self):
        BaseNode.__init__(self)

    def append_lines(self, line_content):
        BaseNode.append_lines(self, line_content)
        self.set_finish_status(self.get_fix_status())


class BodyNode(BaseNode):
    def __init__(self):
        BaseNode.__init__(self)
        self._child_nodes = []

    def add_child_nodes(self, val):
        self._child_nodes.append(val)

    def get_child_nodes(self):
        return self._child_nodes

    def get_fix_status(self):
        if len(self.get_child_nodes()) == 0:
            return False
        for node in self.get_child_nodes():
            if not node.get_fix_status():
                return False
        return True

    def append_lines(self, line_content):
        new_node = get_new_node(self.get_child_nodes())
        if new_node is None:
            if re.match(COMPLEX_NODE_RE_MATCH, line_content, re.IGNORECASE):
                new_node = ComplexNode()
            else:
                new_node = SimpleNode()

            self.add_child_nodes(new_node)

        new_node.append_lines(line_content)

        if get_new_node(self.get_child_nodes()) is not None:
            self.set_finish_status(False)


class ComplexNode(BaseNode):
    def __init__(self):
        BaseNode.__init__(self)
        self._header = HeaderNode()
        self._body = BodyNode()
        self._tail_status = False

    @property
    def arguments(self):
        return self._header._lines

    def get_header(self):
        return self._header

    def get_body(self):
        return self._body

    def get_tail_status(self):
        return self._tail_status

    def get_fix_status(self):
        return self._header.get_fix_status() and self.get_body(
        ).get_finish_status() and self.get_tail_status()

    def append_lines(self, line_content):
        if not self._header.get_finish_status():
            node_body = line_content
            if RIGHT_TIP_SCRATCH in line_content:
                node_body, extra_content = line_content.split(
                    RIGHT_TIP_SCRATCH, 1)

            node_body = node_body.lstrip()
            if node_body.startswith(LEFT_TIP_SCRATCH) and node_body:
                self.set_name(
                    node_body.split(" ")[0].lstrip(LEFT_TIP_SCRATCH).strip())
                self._header.set_name(
                    node_body.split(" ")[0].lstrip(LEFT_TIP_SCRATCH).strip())
                node_body = node_body.lstrip(LEFT_TIP_SCRATCH)
            if node_body:
                self._header.append_lines(node_body)
            if RIGHT_TIP_SCRATCH in line_content:
                self._header.set_finish_status(True)
        elif not self._body.get_fix_status():
            if len(self._body.get_child_nodes()) == 0 and re.match(
                    COMPLEX_NODE_TAIL_RE_MATCH % self._header.get_name(),
                    line_content, re.IGNORECASE):
                self._body.set_finish_status(True)
                self._tail_status = True
            else:
                self._body.append_lines(line_content)
        elif re.match(COMPLEX_NODE_TAIL_RE_MATCH % self._header.get_name(),
                      line_content, re.IGNORECASE):
            self._body.set_finish_status(True)
            self._tail_status = True
        elif not self._body.get_finish_status():
            self._body.append_lines(line_content)


class ApacheConfParser:
    """
    description: The apache config file directive.
    """
    def __init__(self, parse_file):
        """
        Parse the specified file
        :param parse_file: path of parse file
        :return:None
        """
        self._content_nodes = []
        self._source_contents = []
        self._server_root = '/usr/local/apache2'
        self._define_root_value = ''
        self._define_server_root_path = ''
        self.get_server_information(parse_file)

        self.optimize_file_content(parse_file, self._source_contents)
        if len(self._source_contents) == 0:
            return

        apache_contents_parse(self)

    def add_content_nodes(self, val):
        self._content_nodes.append(val)

    def get_source_contents(self):
        return self._source_contents

    def get_content_nodes(self):
        return self._content_nodes

    def get_server_information(self, parse_file):
        """
        Parse and optimize files, and get path of server root
        :param parse_file: path of parse file
        :return:None
        """
        parse_contents = open(parse_file)
        for line_content in parse_contents:
            line_content = line_content.strip()
            if re.match(COMMENT_RE_MATCH,
                        line_content, re.IGNORECASE) or re.match(
                            BLANK_RE_MATCH, line_content, re.IGNORECASE):
                continue

            if re.match(DEFINE_RE_MATCH, line_content, re.IGNORECASE):
                self._define_root_value = ''
                pattern = re.compile(DEFINE_RE_MATCH, re.IGNORECASE)
                server_root_path = pattern.findall(line_content)
                if server_root_path:
                    self._define_root_value = server_root_path[0][0].strip(
                    ).strip('\"')
                    self._define_server_root_path = server_root_path[0][
                        1].strip().strip('\"')
                else:
                    raise Exception(
                        "Directive ServerRoot is error configured!")
            elif re.match(SERVER_ROOT_RE_MATCH, line_content, re.IGNORECASE):
                pattern = re.compile(SERVER_ROOT_RE_MATCH, re.IGNORECASE)
                root_path_list = pattern.findall(line_content)
                if root_path_list:
                    check_value = root_path_list[0].strip().strip('\"')
                    if check_value == '${' + self._define_root_value + '}':
                        self._server_root = self._define_server_root_path
                    else:
                        self._server_root = root_path_list[0].strip().strip(
                            '\"')
                else:
                    raise Exception(
                        "Directive ServerRoot is error configured!")

    def optimize_file_content(self, parse_file, new_contents):
        """
        Parse and optimize files, delete comment and blank line, parse include files
        :param parse_file: path of parse file
        :param new_contents:  line content list
        :return:None
        """
        parse_contents = open(parse_file)
        for line_content in parse_contents:
            line_content = line_content.strip()
            if re.match(COMMENT_RE_MATCH,
                        line_content, re.IGNORECASE) or re.match(
                            BLANK_RE_MATCH, line_content, re.IGNORECASE):
                continue

            if re.match(INCLUDE_RE_MATCH, line_content, re.IGNORECASE):
                pattern = re.compile(INCLUDE_RE_MATCH, re.IGNORECASE)
                include_path_list = pattern.findall(line_content)
                if include_path_list:
                    include_paths = include_path_list[0][-1].strip().strip(
                        '\"').split()
                    for include_path in include_paths:
                        include_path = include_path.strip()
                        self.parse_include_file(include_path, new_contents)
                else:
                    raise Exception("Directive Include is error configured!")
            else:
                new_contents.append(line_content)

    def parse_include_file(self, include_path, new_contents):
        """
        Parse include files
        :param include_path: path of include file
        :param new_contents:  line content list
        :return:None
        """
        abs_path = include_path
        if not os.path.isabs(include_path):
            abs_path = os.path.join(self._server_root, abs_path)

        abs_path = os.path.realpath(abs_path)
        glob_path = glob.glob(abs_path)
        for path_s in glob_path:
            if path_s == '':
                continue
            if os.path.exists(path_s) is False:
                continue
            if os.path.getsize(path_s) <= 0:
                continue

            include_contents = []
            self.optimize_file_content(path_s, include_contents)
            new_contents.extend(include_contents)

    def find_all_directive(self, name):
        """
        Find all nodes according to name
        :param name:name, type is list
        :return:all nodes, type is list
        """
        ret = []

        for node in self.get_content_nodes():
            if isinstance(node, SimpleNode):
                if node.get_name().lower() == name.lower():
                    ret.append(node)
            else:
                ret.extend(
                    find_all_directive_in_directive(
                        node.get_body().get_child_nodes(), name))

        return ret

    def find_directive(self, path):
        """
        Find all nodes according to path
        :param path:path, type is list
        :return:all nodes, type is list
        """
        if path is None:
            path = []

        length = len(path)
        # no path ,return empty list
        if length == 0:
            return []

        ret_value = []
        for node in self.get_content_nodes():
            ret_value.extend(find_directive_in_directive(node, path))

        return ret_value


def get_new_node(nodes):
    """
    Parse file contents
    :param nodes:The object of apache config file directive.
    :return:The newest node, type is SimpleNode or ComplexNode
    """
    for node in nodes:
        if not node.get_fix_status():
            return node

    return None


def apache_contents_parse(apache_conf_object):
    """
    Parse file contents
    :param apache_conf_object:The object of apache config file directive.
    :return:None
    """
    source_contents = apache_conf_object.get_source_contents()
    for line_content in source_contents:
        # Check if newly created node object or append content on old object
        new_node = get_new_node(apache_conf_object.get_content_nodes())
        if new_node is None:
            if re.match(COMPLEX_NODE_RE_MATCH, line_content, re.IGNORECASE):
                new_node = ComplexNode()
            else:
                new_node = SimpleNode()

            apache_conf_object.add_content_nodes(new_node)

        new_node.append_lines(line_content)


def find_all_directive_in_directive(nodes, name):
    """
    Find all nodes according to name
    :param nodes:node, type is SimpleNode or ComplexNode
    :param name:path, type is list
    :return:Directory path, type is list
    """
    ret = []
    for node in nodes:
        if isinstance(node, SimpleNode):
            if node.get_name().lower() == name.lower():
                ret.append(node)
        else:
            ret.extend(
                find_all_directive_in_directive(
                    node.get_body().get_child_nodes(), name))
    return ret


def find_directive_in_directive(node, path):
    """
    Find all nodes according to path
    :param node:node, type is SimpleNode or ComplexNode
    :param path:path, type is list
    :return:Directory path, type is list
    """
    if path is None:
        return []

    length = len(path)
    if length <= 0:
        return []

    ret_value = []
    if isinstance(node, SimpleNode):
        if length == 1:
            if path[0].lower() == node.get_name().lower():
                ret_value.append(node)

    else:
        if length == 1:
            if path[0].lower() == node.get_name().lower():
                ret_value.append(node)
        else:
            if path[0].lower() == node.get_name().lower():
                for node_value in node.get_body().get_child_nodes():
                    ret_value.extend(
                        find_directive_in_directive(node_value, path[1:]))

    return ret_value
